/* msDistri - microscope distributor
 * version 0.1 date 2009-06-19
 * Copyright (C) 2009 Stefan Schuermans <stefan@blinkenarea.org>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 * a BlinkenArea project - http://www.blinkenarea.org/
 */

#include <string.h>
#include "demo.h"
#include "ethernet.h"
#include "ip.h"
#include "layout.h"
#include "mcuf.h"
#include "serial.h"
#include "status.h"
#include "udp.h"

// process received MCUF frame message
static void McufProcFrame( unsigned char * pData, unsigned short Length )
{
  // check header: MCUF <TOTAL_WIDTH>x<TOTAL_HEIGHT>-1/256 (magic is already checked)
  if( Length < 12 )
    return;
  if( pData[4] != 0 || pData[5] != TOTAL_HEIGHT )
    return;
  if( pData[6] != 0 || pData[7] != TOTAL_WIDTH )
    return;
  if( memcmp( pData + 8, "\x00\x01\x00\xFF", 4 ) != 0 )
    return;
  if( Length < 12 + TOTAL_HEIGHT * TOTAL_WIDTH )
    return;
  unsigned char * pixels = pData + 12; // get pointer to pixel data

  // signal MCUF event to status
  StatusEventMcuf( );

  // transmit header on every output
  unsigned char data[8] = {
    UPPER_HEIGHT, UPPER_HEIGHT, UPPER_HEIGHT, UPPER_HEIGHT,
    LOWER_HEIGHT, LOWER_HEIGHT, LOWER_HEIGHT, LOWER_HEIGHT
  };
  SerialOutputAll( 0x23 ); // header
  SerialOutputAll( 0x54 );
  SerialOutputAll( 0x26 );
  SerialOutputAll( 0x66 );
  SerialOutputAll( 0x00 ); // height
  SerialOutputEach( data );
  SerialOutputAll( 0x00 ); // width
  SerialOutputAll( 0x08 );
  SerialOutputAll( 0x00 ); // channels
  SerialOutputAll( 0x01 );
  SerialOutputAll( 0x00 ); // maxval
  SerialOutputAll( 0xFF );

  // transmit pixel data
  unsigned char * upper_src = pixels;
  unsigned char * lower_src = pixels + UPPER_HEIGHT * TOTAL_WIDTH;
  for( unsigned char y = 0; y < UPPER_HEIGHT || y < LOWER_HEIGHT; y++ ) {
    for( unsigned char x = 0; x < 8; x++ ) {
      unsigned char * dest = data;
      for( unsigned char o = 0, sx = x; o < 4; o++, sx += 8, dest++ )
        *dest = sx < TOTAL_WIDTH && y < UPPER_HEIGHT ? upper_src[sx] : 0;
      for( unsigned char o = 0, sx = x; o < 4; o++, sx += 8, dest++ )
        *dest = sx < TOTAL_WIDTH && y < LOWER_HEIGHT ? lower_src[sx] : 0;
      SerialOutputEach( data );
    }
    upper_src += TOTAL_WIDTH;
    lower_src += TOTAL_WIDTH;
  }
}

// process received PWM parameter message
//   - format: <header> <parameters>
//      - <header> := "PARM" <Y (16 bit)> <X (16 bit)>
//      - <parameters> := <parameter 1> ... <parameter 4>
//      - <parameter I> := byte, 0..255, coefficient for channel value ** I
static void McufProcParam( unsigned char * pData, unsigned short Length )
{
  // check header and get information from it: "PARM" [0..15]x[0..23] (magic is already checked)
  if( Length < 8 )
    return;
  if( pData[4] != 0x00 || pData[6] != 0x00 )
    return;
  unsigned char y = pData[5];
  unsigned char x = pData[7];
  if( y >= TOTAL_HEIGHT || x >= TOTAL_WIDTH )
    return;

  // check length of calibration parameters
  if( Length < 8 + 4 )
    return;
  unsigned char * parameters = pData + 8; // get pointer to PWM parameters

  // calculate output to send parameter message on
  unsigned char output = x / 8; // 8 LED strips at one output
  x %= 8; // x coordinate relative to group of 8 LED strips
  if( y >= UPPER_HEIGHT ) {
    y -= UPPER_HEIGHT; // y coordinate relative to upper/lower part
    output += 4; // lower part: outputs 4..7
  }

  // signal MCUF event to status
  StatusEventMcuf( );

  // transmit PWM parameter message on one serial output (and 0x00 on others)
  SerialOutputOne( output, 0x50, 0x00 ); // header
  SerialOutputOne( output, 0x41, 0x00 );
  SerialOutputOne( output, 0x52, 0x00 );
  SerialOutputOne( output, 0x4D, 0x00 );
  SerialOutputOne( output, 0x00, 0x00 ); // Y
  SerialOutputOne( output, y, 0x00 );
  SerialOutputOne( output, 0x00, 0x00 ); // X
  SerialOutputOne( output, x, 0x00 );
  for( unsigned char i = 0; i < 4; i++ ) // PWM parameters
    SerialOutputOne( output, parameters[i], 0x00 );
}

// process received save message
//   - format: <header>
//      - <header> := "SAVE"
static void McufProcSave( unsigned char * pData, unsigned short Length )
{
  // check header (magic is already checked)
  //  - nothing else in header
  if( Length < 4 )
    return;

  // signal MCUF event to status
  StatusEventMcuf( );

  // transmit save message on all serial outputs
  SerialOutputAll( 0x53 ); // header
  SerialOutputAll( 0x41 );
  SerialOutputAll( 0x56 );
  SerialOutputAll( 0x45 );
}

// process received MCUF message
static void McufProcMsg( unsigned char * pData, unsigned short Length )
{
  // get magic
  if( Length < 4 )
    return;
  unsigned long magic = (unsigned long)pData[0] << 24
                      | (unsigned long)pData[1] << 16
                      | (unsigned long)pData[2] << 8
                      | (unsigned long)pData[3];

  // jump to processing according to magic
  switch( magic ) {
    case 0x23542666: // MCUF frame
      McufProcFrame( pData, Length );
      break;
    case 0x5041524D: // PWM parameter message
      McufProcParam( pData, Length );
      break;
    case 0x53415645: // save message
      McufProcSave( pData, Length );
      break;
  }
}

// process a received MCUF packet
void McufRecv( unsigned char * pData, unsigned short Length )
{
  // short packet
  if( Length < sizeof( struct UdpPacket ) )
    return;

  // get pointer to data and length of data
  unsigned char * ptr = pData + sizeof( struct UdpPacket );
  unsigned short len = Length - sizeof( struct UdpPacket );

  // process MCUF message
  McufProcMsg( ptr, len );
}

